/*
 * Copyright 2022
 *   Gal Assa <galassa@campus.technion.ac.il>
 *   Andreia Correia <andreia.veiga@unine.ch>
 *   Pedro Ramalhete <pramalhe@gmail.com>
 *   Valerio Schiavoni <valerio.schiavoni@unine.ch>
 *   Pascal Felber <pascal.felber@unine.ch>
 *
 * This work is published under the MIT license. See LICENSE.txt
 */

#pragma once

#include <atomic>
#include <cassert>
#include <functional>
#include <cstring>
#include <thread>       // Needed by this_thread::yield()
#include <sys/mman.h>   // Needed if we use mmap()
#include <sys/types.h>  // Needed by open() and close()
#include <sys/stat.h>
#include <linux/mman.h> // Needed by MAP_SHARED_VALIDATE
#include <fcntl.h>
#include <unistd.h>     // Needed by close()
#include <filesystem>   // Needed by std::filesystem::space()
#include <type_traits>
#include <sched.h>      // sched_setaffinity()
#include <csetjmp>      // Needed by sigjmp_buf
#include <cstdlib>      // Needed by exit()
#include <iostream>
#include <chrono>
#include <string>
#include <stdio.h>


/*
 * <h1> TL4x </h1>
 *
 * We use 4 regions, two volatile and two persistent.
 * In this implementation we place the two volatile regions one after the other in a single volatile mmap and
 * the two persistent regions one after the other in a single persistent mmap. This means we have two files,
 * one of the file is in /dev/shm (or tmpfs) and the other file is in a PM DAX filesystem (or just on disk).
 * The volatile regions are referred to as 'main' and 'back', while the persistent regions are 'p0' and 'p1'.
 *
 * Each region is composed of the PMetadata (256 bytes) followed by the EsLocoTu metadata block and then
 * the user-allocated objects (managed by the EslocoTu allocator):
 * +-----------+------------+------------------------+
 * | PMetadata | ELMetadata | User-allocated objects |
 * +-----------+------------+------------------------+
 *
 *
 */


// Macros needed for persistence
#ifdef PWB_IS_CLFLUSH
  /*
   * More info at http://elixir.free-electrons.com/linux/latest/source/arch/x86/include/asm/special_insns.h#L213
   * Intel programming manual at https://www.intel.com/content/dam/www/public/us/en/documents/manuals/64-ia-32-architectures-optimization-manual.pdf
   * Use these for Broadwell CPUs (cervino server)
   */
  #define PWB(addr)                 __asm__ volatile("clflush (%0)" :: "r" (addr) : "memory")                      // Broadwell only works with this.
  #define PFENCE()                  {}                                                                             // No ordering fences needed for CLFLUSH (section 7.4.6 of Intel manual)
  #define PSYNC()                   {}                                                                             // For durability it's not obvious, but CLFLUSH seems to be enough, and PMDK uses the same approach
  #define MSYNC(_addr, _len, _flag) {}
#elif PWB_IS_CLWB
  /* Use this for CPUs that support clwb, such as the SkyLake SP series (c5 compute intensive instances in AWS are an example of it) */
  #define PWB(addr)                 __asm__ volatile(".byte 0x66; xsaveopt %0" : "+m" (*(volatile char *)(addr)))  // clwb() only for Ice Lake onwards
  #define PFENCE()                  __asm__ volatile("sfence" : : : "memory")
  #define PSYNC()                   __asm__ volatile("sfence" : : : "memory")
  #define MSYNC(_addr, _len, _flag) {}
#elif PWB_IS_NOP
  /* pwbs are not needed for shared memory persistency (i.e. persistency across process failure) */
  #define PWB(addr)                 {}
  #define PFENCE()                  __asm__ volatile("sfence" : : : "memory")
  #define PSYNC()                   __asm__ volatile("sfence" : : : "memory")
  #define MSYNC(_addr, _len, _flag) {}
#elif PWB_IS_CLFLUSHOPT
  /* Use this for CPUs that support clflushopt, which is most recent x86 */
  #define PWB(addr)                 __asm__ volatile(".byte 0x66; clflush %0" : "+m" (*(volatile char *)(addr)))    // clflushopt (Kaby Lake)
  #define PFENCE()                  __asm__ volatile("sfence" : : : "memory")
  #define PSYNC()                   __asm__ volatile("sfence" : : : "memory")
  #define MSYNC(_addr, _len, _flag) {}
#elif PWB_IS_DISK
  /* Use this when the mmap is backed by a file on disk */
  #define PWB(addr)                 {}
  #define PFENCE()                  {}
  #define PSYNC()                   {}
  #define MSYNC(_addr, _len, _flag) ::msync(_addr, _len, _flag)
#else
#error "You must define what PWB is. Choose PWB_IS_CLFLUSHOPT if you don't know what your CPU is capable of"
#endif


namespace tl4x {

//
// User configurable variables.
// Feel free to change these if you need larger transactions, or more threads.
//

// Turn this one to see when the copy-thread executes a SYNC operation
#define DEBUG_SYNC false

// Start address of persistent memory region
#ifndef PM_REGION_BEGIN
#define PM_REGION_BEGIN 0x7fea00000000
#endif
// Size of the persistent memory region
#ifndef PM_REGION_SIZE
#define PM_REGION_SIZE 1024*1024*1024ULL // 1GB for now
#endif
// Name of persistent file mapping (back)
#ifndef PM_FILE_NAME
#define PM_FILE_NAME   "/dev/shm/tl4x_shared"
#endif

// Maximum number of registered threads that can execute transactions
static const int REGISTRY_MAX_THREADS = 256;    // An AWS c6a.48xlarge instance has 192 vCPUs, let's make sure we can run there.
// Maximum number of root pointers available for the user
static const uint64_t MAX_ROOT_POINTERS = 64;

static const char* VR_FILE_NAME = "/dev/shm/tl4x_shared_main"; // TODO: try placing this on the disk to see if it impacts performance
// Start address of volatile memory region (VR)
static uint8_t* VR_REGION_ADDR = (uint8_t*)0x7f0000000000;

// Important addresses
static const uint64_t MAIN_REGION_SIZE = PM_REGION_SIZE/2;
static uint8_t* const BEGIN_MAIN_REGION = VR_REGION_ADDR;
static uint8_t* const END_MAIN_REGION = BEGIN_MAIN_REGION + MAIN_REGION_SIZE;
static uint8_t* const BEGIN_BACK_REGION = BEGIN_MAIN_REGION + MAIN_REGION_SIZE;
static uint8_t* const BEGIN_P0_REGION = (uint8_t*)PM_REGION_BEGIN;
static uint8_t* const BEGIN_P1_REGION = (uint8_t*)PM_REGION_BEGIN + PM_REGION_SIZE/2;

// How many write transactions we execute before attempting a SYNC
static const uint64_t MIN_TXN_BETWEEN_SYNC = 10*1000ULL; // 10K txn
//static const uint64_t MIN_TXN_BETWEEN_SYNC = 1000ULL;  // one SYNC every 1k txn. Enable this when running the stress-crash tests

// TL2 constants
static const uint64_t WRITE_LOCKED = 0x8000000000000000ULL;
static const int TX_IS_NONE   = 0;
static const int TX_IS_UPDATE = 2;
static const int TX_IS_READ   = 1;
static const int TX_IS_IRR_READ = 3;
static const int TX_IS_SNAPSHOT_READ = 4;


// Number of striped locks. Each lock protects one 64 bytes. MUST be a power of 2.
static const uint64_t NUM_LOCKS = 32*1024*1024;
// Size of the data that each lock protects in number of bits. 6 bits means 2^6 = 64 bytes
static const uint64_t BIT_SHIFT_HASH = 6;

// There are four possible states for gState:
static const uint8_t STATE_DUPLICATE       = 0;  // The write transactions duplicate the changes on the 'back'
static const uint8_t STATE_PERSIST         = 1;  // The write transactions do not touch the 'back' because the copy-thread needs a snapshot to persist the data
static const uint8_t STATE_READ_SNAPSHOT   = 2;  // The write transactions: same as above. Snapshot and Irrevocable readers may perform reads on 'back'
static const uint8_t STATE_DUPLICATE_BLOCK = 3;  // The write transactions duplicate on 'back' the entire block of every acquired lock, even if they abort the transaction
static const uint64_t COUNTER_ON = 0x8000000000000000ULL; // counter can be incremented

// TODO: remame following. Used for (a) comparing counter to 0; and (2) reset state (duplicate-block to duplicate)
static const uint64_t NO_INFLIGHT_READERS = 0x3FFFFFFFFFFFFFFF; // TODO: 0x1FFF... if there's a fifth state

#define GETSTATE(x_) (x_ >> 62) % 4

// Function that hashes every 64 bytes of data into a lock
inline static uint64_t hidx(void* addr) {
    return (((size_t)addr >> BIT_SHIFT_HASH) & (NUM_LOCKS-1));
}
// Helper functions
inline static bool isLocked(uint64_t sl) { return (sl & WRITE_LOCKED); }
// Helper method: sl is unlocked and lower than the clock, or it's locked by me
static inline bool isUnlockedOrLockedByMe(uint64_t rClock, uint64_t tid, uint64_t sl) {
    return ( (!(sl & WRITE_LOCKED) && (sl <= rClock)) || (sl == (tid|WRITE_LOCKED)) );
}


//
// Thread Registry stuff
//
extern void thread_registry_deregister_thread(const int tid);

// An helper class to do the checkin and checkout of the thread registry
struct ThreadCheckInCheckOut {
    static const int NOT_ASSIGNED = -1;
    int tid { NOT_ASSIGNED };
    ~ThreadCheckInCheckOut() {
        if (tid == NOT_ASSIGNED) return;
        thread_registry_deregister_thread(tid);
    }
};

extern thread_local ThreadCheckInCheckOut tl_tcico;

// Forward declaration of global/singleton instance
class ThreadRegistry;
extern ThreadRegistry gThreadRegistry;

/*
 * <h1> Registry for threads </h1>
 *
 * This is singleton type class that allows assignement of a unique id to each thread.
 * The first time a thread calls ThreadRegistry::getTID() it will allocate a free slot in 'usedTID[]'.
 * This tid wil be saved in a thread-local variable of the type ThreadCheckInCheckOut which
 * upon destruction of the thread will call the destructor of ThreadCheckInCheckOut and free the
 * corresponding slot to be used by a later thread.
 */
class ThreadRegistry {
private:
    alignas(128) std::atomic<bool>      usedTID[REGISTRY_MAX_THREADS];   // Which TIDs are in use by threads
    alignas(128) std::atomic<int>       maxTid {-1};                     // Highest TID (+1) in use by threads

public:
    ThreadRegistry() {
        for (int it = 0; it < REGISTRY_MAX_THREADS; it++) {
            usedTID[it].store(false, std::memory_order_relaxed);
        }
    }

    // Progress condition: wait-free bounded (by the number of threads)
    int register_thread_new(void) {
        for (int tid = 0; tid < REGISTRY_MAX_THREADS; tid++) {
            if (usedTID[tid].load(std::memory_order_acquire)) continue;
            bool unused = false;
            if (!usedTID[tid].compare_exchange_strong(unused, true)) continue;
            // Increase the current maximum to cover our thread id
            int curMax = maxTid.load();
            while (curMax <= tid) {
                maxTid.compare_exchange_strong(curMax, tid+1);
                curMax = maxTid.load();
            }
            tl_tcico.tid = tid;
            return tid;
        }
        printf("ERROR: Too many threads, registry can only hold %d threads\n", REGISTRY_MAX_THREADS);
        assert(false);
        return 0;
    }

    // Progress condition: wait-free population oblivious
    inline void deregister_thread(const int tid) {
        usedTID[tid].store(false, std::memory_order_release);
    }

    // Progress condition: wait-free population oblivious
    static inline uint64_t getMaxThreads(void) {
        return gThreadRegistry.maxTid.load(std::memory_order_acquire);
    }

    // Progress condition: wait-free bounded (by the number of threads)
    static inline int getTID(void) {
        int tid = tl_tcico.tid;
        if (tid != ThreadCheckInCheckOut::NOT_ASSIGNED) return tid;
        return gThreadRegistry.register_thread_new();
    }
};


/*
 * EsLocoTu? is an Extremely Simple memory aLOCatOr Two
 *
 * It is based on intrusive singly-linked stacks (a free-list), one for each power of two size.
 * All blocks are powers of two, the smallest size enough to contain the desired user data plus the block header.
 * There is an array named 'freelists' where each entry is a pointer to the head of a stack for that respective block size.
 * Blocks are allocated in powers of 2 of words (64bit words).
 * Each block has an header with two words: the size of the node (in words), the pointer to the next node.
 * The minimum block size is 4 words, with 2 for the header and 2 for the user.
 * When there is no suitable block in the freelist, it will create a new block from the remaining pool.
 *
 * EsLoco was designed for usage in PTMs but it doesn't have to be used only for that.
 * EsLocoTu scales while EsLoco does not.
 * Average number of stores for an allocation is 1.
 * Average number of stores for a de-allocation is 2.
 *
 * Memory layout:
 * ---------------------------------------------------------------------------------------------------
 * | poolTop | gfreelists[0..50] | threadflists[0..REGISTRY_MAX_THREADS] | ... allocated objects ... |
 * ---------------------------------------------------------------------------------------------------
 */
template <template <typename> class P>
class EsLoco2 {
private:
    const bool debugOn = false;
    // Number of blocks in the freelists array.
    // Each entry corresponds to an exponent of the block size: 2^4, 2^5, 2^6... 2^40
    static const int kMaxBlockSize = 50; // 1024 TB of memory should be enough
    // Size after which we trigger a cleanup of the thread-local list/pool
    static const uint64_t kMaxListSize = 64;
    // Size of a thread-local slab (in bytes). When it runs out, we take another slab from poolTop.
    // Larger allocations go directly on poolTop which means doing a lot of large allocations from
    // multiple threads will result in contention.
    static const uint64_t kSlabSize = 4*1024*1024;

    struct block {
        P<block*>   next;    // Pointer to next block in free-list (when block is in free-list)
        P<uint64_t> size2;   // Exponent of power of two of the size of this block in bytes.
    };

    // A persistent free-list of blocks
    struct FList {
        P<uint64_t> count;   // Number of blocks in the stack
        P<block*>   head;    // Head of the stack (where we push and pop)
        P<block*>   tail;    // Tail of the stack (used for faster transfers to/from gflist)

        // Push a block into this free list, incrementing the volatile
        // counter and adjusting the tail if needed. Does two pstores min.
        void pushBlock(block* myblock) {
            block* lhead = head;
            if (lhead == nullptr) {
                assert(count == 0);
                tail = myblock;                    // pstore()
            }
            myblock->next = lhead;                 // pstore()
            head = myblock;                        // pstore()
            count++;
        }

        // Pops a block from a given (persistent) free list, decrementing the
        // volatile counter and setting the tail to nullptr if list becomes empty.
        // Does 2 pstores.
        block* popBlock(void) {
            assert(count != 0);
            assert(tail != nullptr);
            count--;
            block* myblock = head;
            head = myblock->next;                // pstore()
            if (head == nullptr) tail = nullptr; // pstore()
            return myblock;
        }

        bool isFull(void) { return (count >= kMaxListSize); }
        bool isEmpty() { return (count == 0); }

        // Transfer this list to another free-list (oflist).
        // We use this method to transfer from the thread-local list to the global list.
        void transfer(FList* oflist) {
            assert(count != 0);
            assert(tail != nullptr);
            if (oflist->tail == nullptr) {
                assert (oflist->head == nullptr);
                oflist->tail = tail;                           // pstore()
            }
            tail->next = oflist->head;                         // pstore()
            oflist->head = head;                               // pstore()
            oflist->count += count;                            // pstore()
            count = 0;                                         // pstore()
            tail = nullptr;                                    // pstore()
            head = nullptr;                                    // pstore()
        }
    };

    // To avoid false sharing and lock conflicts, we put each thread-local data in one of these.
    // This is persistent (in PM).
    struct TLData {
        uint64_t     padding[8];
        // Thread-local array of persistent heads of free-list
        FList        tflists[kMaxBlockSize];
        // Thread-local slab pointer
        P<uint8_t*>  slabtop;
        // Thread-loal slab size
        P<uint64_t>  slabsize;
        TLData() {
            for (int i = 0; i < kMaxBlockSize; i++) {
                tflists[i].count = 0;             // pstore()
                tflists[i].head = nullptr;        // pstore()
                tflists[i].tail = nullptr;        // pstore()
            }
            slabtop = nullptr;                    // pstore()
            slabsize = 0;                         // pstore()
        }
    };

    // Complete metadata. This is PM
public:
    struct ELMetadata {
        P<uint8_t*> top;
        FList       gflists[kMaxBlockSize];
        TLData      tldata[REGISTRY_MAX_THREADS];
        // This constructor should be called from within a transaction
        ELMetadata() {
            // The 'top' starts after the metadata header
            top = ((uint8_t*)this) + sizeof(ELMetadata);         // pstore()
            for (int i = 0; i < kMaxBlockSize; i++) {
                gflists[i].count = 0;             // pstore()
                gflists[i].head = nullptr;        // pstore()
                gflists[i].tail = nullptr;        // pstore()
            }
            for (int it = 0; it < REGISTRY_MAX_THREADS; it++) {
                new (&tldata[it]) TLData();                      // pstores()
            }
        }
    };

    // Volatile data
    uint8_t* poolAddr {nullptr};
    uint64_t poolSize {0};

    // Pointer to location of PM metadata of EsLoco
    ELMetadata* meta {nullptr};

    // For powers of 2, returns the highest bit, otherwise, returns the next highest bit
    uint64_t highestBit(uint64_t val) {
        uint64_t b = 0;
        while ((val >> (b+1)) != 0) b++;
        if (val > (1ULL << b)) return b+1;
        return b;
    }

    uint8_t* aligned(uint8_t* addr) {
        return (uint8_t*)((size_t)addr & (~0x3FULL)) + 128;
    }

public:
    void init(void* addressOfMemoryPool, size_t sizeOfMemoryPool, bool clearPool=true) {
        // Align the base address of the memory pool
        poolAddr = aligned((uint8_t*)addressOfMemoryPool);
        poolSize = sizeOfMemoryPool + (uint8_t*)addressOfMemoryPool - poolAddr;
        // The first thing in the pool is the ELMetadata
        meta = (ELMetadata*)poolAddr;
        if (clearPool) new (meta) ELMetadata();
        if (debugOn) printf("Starting EsLoco2 with poolAddr=%p and poolSize=%ld, up to %p\n", poolAddr, poolSize, poolAddr+poolSize);
    }

    // Resets the metadata of the allocator back to its defaults
    void reset() {
        std::memset(poolAddr, 0, sizeof(block)*kMaxBlockSize*(REGISTRY_MAX_THREADS+1));
        meta->top.pstore(nullptr);
    }

    // Returns the number of bytes that may (or may not) have allocated objects, from the base address to the top address
    uint64_t getUsedSize() {
        return meta->top.pload() - poolAddr;
    }

    // Takes the desired size of the object in bytes.
    // Returns pointer to memory in pool, or nullptr.
    // Does on average 1 store to persistent memory when re-utilizing blocks.
    void* malloc(size_t size) {
        const int tid = ThreadRegistry::getTID();
        // Adjust size to nearest (highest) power of 2
        uint64_t bsize = highestBit(size + sizeof(block));
        uint64_t asize = (1ULL << bsize);
        if (debugOn) printf("malloc(%ld) requested,  block size exponent = %ld\n", size, bsize);
        block* myblock = nullptr;
        // Check if there is a block of that size in the corresponding freelist
        if (!meta->tldata[tid].tflists[bsize].isEmpty()) {
            if (debugOn) printf("Found available block in thread-local freelist\n");
            myblock = meta->tldata[tid].tflists[bsize].popBlock();
        } else if (asize < kSlabSize/2) {
            // Allocations larger than kSlabSize/2 go on the global top
            if (asize >= meta->tldata[tid].slabsize.pload()) {
                if (debugOn) printf("Not enough space on current slab of thread %d for %ld bytes. Taking new slab from top\n",tid, asize);
                // Not enough space for allocation. Take a new slab from the global top and
                // discard the old one. This means fragmentation/leak is at most kSlabSize/2.
                if (meta->top.pload() + kSlabSize > poolSize + poolAddr) {
                    printf("EsLoco2: Out of memory for slab %ld for %ld bytes allocation\n", kSlabSize, size);
                    return nullptr;
                }
                meta->tldata[tid].slabtop = meta->top.pload();   // pstore()
                meta->tldata[tid].slabsize = kSlabSize;          // pstore()
                meta->top.pstore(meta->top.pload() + kSlabSize); // pstore()
                // TODO: add the remains of the slab to thread-local freelist
            }
            // Take space from the slab
            myblock = (block*)meta->tldata[tid].slabtop.pload();
            myblock->size2 = bsize;                                      // pstore()
            meta->tldata[tid].slabtop.pstore((uint8_t*)myblock + asize); // pstore()
            meta->tldata[tid].slabsize -= asize;                         // pstore()
        } else if (!meta->gflists[bsize].isEmpty()) {
            if (debugOn) printf("Found available block in thread-global freelist\n");
            myblock = meta->gflists[bsize].popBlock();
        } else {
            if (debugOn) printf("Creating new block from top, currently at %p\n", meta->top.pload());
            // Couldn't find a suitable block, get one from the top of the pool if there is one available
            if (meta->top.pload() + asize > poolSize + poolAddr) {
                printf("EsLoco2: Out of memory for %ld bytes allocation\n", size);
                return nullptr;
            }
            myblock = (block*)meta->top.pload();
            meta->top.pstore(meta->top.pload() + asize);                 // pstore()
            myblock->size2 = bsize;                                       // pstore()
        }
        if (debugOn) printf("returning ptr = %p\n", (void*)((uint8_t*)myblock + sizeof(block)));
        //printf("malloc:%ld = %p\n", asize, (void*)((uint8_t*)myblock + sizeof(block)));
        // Return the block, minus the header
        return (void*)((uint8_t*)myblock + sizeof(block));
    }

    // Takes a pointer to an object and puts the block on the free-list.
    // When the thread-local freelist becomes too large, all blocks are moved to the thread-global free-list.
    // Does on average 2 stores to persistent memory.
    void free(void* ptr) {
        if (ptr == nullptr) return;
        const int tid = ThreadRegistry::getTID();
        block* myblock = (block*)((uint8_t*)ptr - sizeof(block));
        if (debugOn) printf("free:%ld %p\n", 1UL << myblock->size2.pload(), ptr);
        // Insert the block in the corresponding freelist
        uint64_t bsize = myblock->size2;
        meta->tldata[tid].tflists[bsize].pushBlock(myblock);
        if (meta->tldata[tid].tflists[bsize].isFull()) {
            if (debugOn) printf("Moving thread-local flist for bsize=%ld to thread-global flist\n", bsize);
            // Move all objects in the thread-local list to the corresponding thread-global list
            meta->tldata[tid].tflists[bsize].transfer(&meta->gflists[bsize]);
        }
    }
};


// Needed by our benchmarks and to prevent destructor from being called automatically on abort/retry
struct tmbase {
    ~tmbase() { }
};


struct ReadSet {
    struct ReadSetEntry {
        std::atomic<uint64_t>* mBeg;
        std::atomic<uint64_t>* mEnd;
    };

    // We pre-allocate a read-set with this many entries and if more are needed,
    // we grow them dynamically during the transaction and then delete them when
    // reset() is called on beginTx().
    static const int64_t MAX_ENTRIES = 16*1024;

    ReadSetEntry  entries[MAX_ENTRIES];
    uint64_t      size {0};          // Number of loads in the readSet for the current transaction
    ReadSet*      next {nullptr};

    inline void reset() {
        if (next != nullptr) {
            next->reset();   // Recursive call on the linked list of logs
            delete next;
            next = nullptr;
        }
        size = 0;
    }

    inline bool validate(uint64_t rClock, uint64_t tid) {
        for (uint64_t i = 0; i < size; i++) {
            for (std::atomic<uint64_t>* mutex = entries[i].mBeg; mutex <= entries[i].mEnd; mutex++) {
                uint64_t sl = mutex->load(std::memory_order_acquire);
                if (!isUnlockedOrLockedByMe(rClock, tid, sl)) return false;
            }
        }
        if (next != nullptr) return next->validate(rClock, tid); // Recursive call to validate()
        return true;
    }

    inline void add(std::atomic<uint64_t>* mBeg, std::atomic<uint64_t>* mEnd) {
        if (size != MAX_ENTRIES) {
            entries[size].mBeg = mBeg;
            entries[size].mEnd = mEnd;
            size++;
        } else {
            // The current node is full, therefore, search a vacant node or create a new one
            if (next == nullptr) next = new ReadSet();
            next->add(mBeg, mEnd);    // Recursive call to add()
        }
    }
};


// The persistent metadata is a 'header'.
struct PMetadata {
    static const uint64_t   MAGIC_ID = 0x1337bab9;
    // Elsewhere in the code we rely on the fact that the first word MUST be the p_timestamp
    volatile uint64_t       p_timestamp;
    void*                   root {nullptr}; // Immutable once assigned
    uint64_t                id {0};
    // Add padding so we have 16 words in total, which is 256 bytes, the size of an Optane block
    uint64_t                paddinga[16-3];

    PMetadata() {
        p_timestamp = 0;
        PWB(&p_timestamp);
    }
};


struct MainBackSeq {
    uint64_t main;  // Sequence/timestamp for a data block in 'main'
    uint64_t back;  // Sequence/timestamp for a data block in 'back'
};


// Address of Persistent Metadata.
// This relies on PMetadata being the first thing in the region.
static PMetadata* const pmd = (PMetadata*)BEGIN_MAIN_REGION;

// We put the array of locks outside the PTM because we want to access it from stand-alone static methods
extern std::atomic<uint64_t> *gHashLock;
// We combine the two arrays of sequences into a single array for better performance/locality
extern MainBackSeq *gSeqArray;



// Volatile log (write-set)
struct WriteSet {

    struct WriteSetEntry {
        void*     addr;
        uint8_t*  data;  // old value
        uint32_t  length;
    };

    int64_t         maxEntries = 512;          // Max number of log entries available in 'entries'
    int64_t         maxScratch = 4*maxEntries; // Max number of bytes available in 'scratch'
    int64_t         numEntries = 0;            // Number of entries used in 'entries'
    int64_t         numScratch = 0;            // Number of bytes used in 'scratch'
    WriteSetEntry*  entries;
    uint8_t*        scratch;

    WriteSet() {
        entries = new WriteSetEntry[maxEntries];
        scratch = new uint8_t[maxScratch];
    }

    ~WriteSet() {
        delete[] entries;
        delete[] scratch;
    }

    // Increase the size of the arrays by 4x, then copy everything to the new arrays.
    // Yes, it will be slow, but it should be a rare occurrence. Don't inline.
    void growEntries4x(void) {
        maxEntries *= 4;
        WriteSetEntry* newEntries = new WriteSetEntry[maxEntries];
        std::memcpy(newEntries, entries, numEntries*sizeof(WriteSetEntry));
        delete[] entries;
        entries = newEntries;
    }
    void growScratch4x(uint64_t length) {
        while (numScratch+length >= maxScratch) maxScratch *= 4;
        uint8_t* newScratch = new uint8_t[maxScratch];
        std::memcpy(newScratch, scratch, numScratch*sizeof(uint8_t));
        delete[] scratch;
        scratch = newScratch;
    }

    inline void reset() {
        numEntries = 0;
        numScratch = 0;
    }

    // Adds range and old data to the log
    inline void addEntry(void* addr, uint32_t length) {
        if (numEntries == maxEntries) growEntries4x();
        if (numScratch+length >= maxScratch) growScratch4x(length);
        entries[numEntries].addr = addr;
        entries[numEntries].length = length;
        entries[numEntries].data = nullptr; // Not really needed, but easier to debug if something goes wrong
        numEntries++;
    }

    // Adds range and old data to the log
    inline void addData(void* addr, uint32_t length) {
        if (length <= sizeof(uint64_t)) {
            uint64_t oldVal = ((std::atomic<uint64_t>*)addr)->load(std::memory_order_acquire);
            entries[numEntries-1].data = (uint8_t*)oldVal;
        } else {
            // It's a range and doesn't fit in 64 bits, therefore, save it in the oldData buffer
            std::memcpy(&scratch[numScratch], addr, length);
            entries[numEntries-1].data = &scratch[numScratch];
            numScratch += length;
        }
    }

    // Helper function for rollback()
    inline bool rangeIsLockedByMe(void* addr, uint32_t length, uint64_t tid) {
        std::atomic<uint64_t>* mBeg  = &gHashLock[hidx(addr)];
        std::atomic<uint64_t>* mEnd  = &gHashLock[hidx(((uint8_t*)addr) + length-1)];
        for (std::atomic<uint64_t>* mutex = mBeg; mutex <= mEnd; mutex++) {
            uint64_t sl = mutex->load(std::memory_order_acquire);
            if (sl != (WRITE_LOCKED | tid)) return false;
        }
        return true;
    }

    // Rollback a single entry in 'main'. Used when the transaction aborts.
    inline void rollbackSingleEntry(int32_t i) {
        WriteSetEntry* entry = &entries[i];
        if (entry->length <= sizeof(uint64_t)) {
            std::memcpy(entry->addr, &entry->data, entry->length);
        } else {
            std::memcpy(entry->addr, entry->data, entry->length);
        }
    }

    // Rollback modifications: called from abortTx() to revert changes.
    void rollbackMain(uint64_t tid) {
        if (numEntries == 0) return;
        // The last entry in the log is special: we don't revert it unless
        // we have all the locks protecting it. The reason being that maybe the
        // abort was triggered due to failed lock acquisition and therefore,
        // reverting it would be incorrect (is already locked by another tx).
        if (rangeIsLockedByMe(entries[numEntries-1].addr, entries[numEntries-1].length, tid)) rollbackSingleEntry(numEntries-1);
        for (int32_t i=numEntries-2; i >= 0; i--) rollbackSingleEntry(i);
    }

    // Unlock all the locks acquired by this thread
    void unlock(uint64_t nextClock, const uint8_t dupState, const bool isCommittedTxn, uint64_t tid) {
        const uint64_t dataSize = (1 << BIT_SHIFT_HASH);
        for (uint64_t i = 0; i < numEntries; i++) {
            void* addr = entries[i].addr;
            int32_t length = entries[i].length;
            // Handle ranges correctly
            const int64_t iBeg = hidx(addr);
            const int64_t iEnd = hidx(((uint8_t*)addr) + length-1);
            uint64_t offset = 0;
            for (int64_t imut = iBeg; imut <= iEnd; imut++) {
                if (gHashLock[imut].load(std::memory_order_relaxed) == (WRITE_LOCKED | tid)) {

                	if (isCommittedTxn) {
                	    gSeqArray[imut].main = nextClock;
                	}
                    // We don't touch the sequence arrays when the transaction aborts
                    if (dupState == STATE_DUPLICATE_BLOCK //){
                    		&& gSeqArray[imut].back != gSeqArray[imut].main) {
                        // If it's STATE_DUPLCIATE_BLOCK we need to help iff the sequences in main and back don't match.
                        // We help by copying the data block, regardless of the transaction committing or aborting.
                        // Copy the dataSized block from 'main' to the 'back' region
                        uint8_t* blockAddr = (uint8_t*)((((size_t)addr) >> BIT_SHIFT_HASH) << BIT_SHIFT_HASH) + offset;
                        std::memcpy(blockAddr + MAIN_REGION_SIZE, blockAddr, dataSize);
                        // Set the new sequence for 'back' and release the lock with the same sequence (we didn't change anything in 'main')
                        gSeqArray[imut].back = gSeqArray[imut].main;
                    }
                    /*
                    if (isCommittedTxn) {
                        gSeqArray[imut].main = nextClock;
                        // Update the 'back' sequence if we're in DUPLICATE or DUPLICATE_BLOCK
                        if (dupState != STATE_PERSIST && dupState != STATE_READ_SNAPSHOT ) gSeqArray[imut].back = nextClock;
                    }*/
                    gHashLock[imut].store(nextClock, std::memory_order_release);
                }
                offset += dataSize;
            }
        }
    }

    // Duplicate modifications on 'back' region. This is called when the write transaction sees the state DUPLICATE
    inline void duplicateChangesOnBack() {
        for (int32_t i = 0; i < numEntries; i++) {
            std::memcpy((uint8_t*)entries[i].addr + MAIN_REGION_SIZE, entries[i].addr, entries[i].length);
        }
    }

    // Duplicate data blocks on 'back' region. This is called when the write transaction sees the state DUPLICATE_BLOCK
    // TODO: add optimization that sets to nextClock the corresponding blocks in gSeqArray[imut].back and skips the ones that
    // match nextClock, so that we don't keep copying the same block over and over again.
    inline void duplicateBlocksOnBack() {
        const uint64_t dataSize = (1 << BIT_SHIFT_HASH);
        for (int32_t i = 0; i < numEntries; i++) {
            uint8_t* beginAddr = (uint8_t*)((((size_t)entries[i].addr) >> BIT_SHIFT_HASH) << BIT_SHIFT_HASH);
            uint8_t* endAddr = (uint8_t*)((((size_t)entries[i].addr + entries[i].length + dataSize - 1) >> BIT_SHIFT_HASH) << BIT_SHIFT_HASH);
            std::memcpy(beginAddr + MAIN_REGION_SIZE, beginAddr, (endAddr - beginAddr));
        }
    }
};


// URCU, used by the copy-thread to know when is it safe to copy from 'back' to persistent replica p0/p1
class URCU {

private:
    // Cache line padding on x86 covers two cache lines to avoid false sharing
    static const int        CLPAD = 128/sizeof(std::atomic<uint64_t>);
    // Special epoch that means thread is not participating in a transaction
    static const uint64_t   NOT_READING = 0xFFFFFFFFFFFFFFFEULL;
    // ReadIndicator, with padding to avoid false sharing
    std::atomic<uint64_t>   ri[REGISTRY_MAX_THREADS*CLPAD];
    // URCU's clock
    alignas(128) std::atomic<uint64_t> clock {1};

public:
    URCU() {
        for (int itid = 0; itid < REGISTRY_MAX_THREADS; itid++) {
            ri[itid*CLPAD].store(NOT_READING, std::memory_order_relaxed);
        }
    }

    // Equivalent to rcu_read_lock(). Called by TL2 write-txn, but not read-txn.
    inline void arrive(const int tid) {
        uint64_t epoch = clock.load();
        ri[tid*CLPAD].exchange(epoch); // exchg is faster than seq-cst store on x86
    }

    // Equivalent to rcu_read_unlock(). Called by TL2 write-txn, but not read-txn.
    inline void depart(const int tid) {
        ri[tid*CLPAD].store(NOT_READING, std::memory_order_release);
    }

    // Equivalent to rcu_synchronize(). Called by copy-thread
    void synchronize() {
        const uint64_t grace = clock.fetch_add(1);
        while (true) {
            uint64_t oldest = NOT_READING;
            const unsigned maxThreads = ThreadRegistry::getMaxThreads();
            for (unsigned it = 0; it < maxThreads; it++) {
                uint64_t epoch = ri[it*CLPAD].load(std::memory_order_relaxed);
                if (epoch < oldest) oldest = epoch;
            }
            if (oldest > grace) return;
            std::this_thread::yield();
        }
    }
};


// Thread-local data
struct OpData {
    std::jmp_buf env;
    uint64_t     attempt {0};
    uint64_t     tid;
    uint64_t     rClock {0};
    int          tx_type {TX_IS_NONE}; // This is used by persist::load() to figure out if it needs to save a load on the read-set or not
    ReadSet      readSet;              // The (volatile) read set
    WriteSet     writeSet {};          // Write-set: The append-only log of modified persist<T>
    uint64_t     myrand;
    uint64_t     numAborts {0};
    uint64_t     numCommits {0};
    uint64_t     state {0};
    uint64_t     padding[16];
};

extern std::atomic<uint64_t> gClock;
extern std::atomic<uint64_t> gState;
extern std::atomic<uint64_t> gCounter;

[[noreturn]] extern void abortTx(OpData* myd);

// This is used by addToLog() to know which OpData instance to use for the current transaction
extern thread_local OpData* tl_opdata;

// Helper function to lock an entire range. Used by pstore() and some of the string utils
inline static void logLockRange(void* vraddr, int32_t length) {
    OpData* const myd = tl_opdata;
    // We must log _before_ locking because, in case of an abort half-way through
    // the lock acquisitions on a range, we want to revert those acquisitions and
    // for that, we need to have that range kept already in the log.
    myd->writeSet.addEntry(vraddr, length);
    std::atomic<uint64_t>* mBeg  = &gHashLock[hidx(vraddr)];
    std::atomic<uint64_t>* mEnd  = &gHashLock[hidx((uint8_t*)vraddr + length-1)];
    for (std::atomic<uint64_t>* mutex = mBeg; mutex <= mEnd; mutex++) {
        uint64_t sl = mutex->load(std::memory_order_acquire);
        if (!isUnlockedOrLockedByMe(myd->rClock, myd->tid, sl)) abortTx(myd);
        if (!isLocked(sl) && !mutex->compare_exchange_strong(sl, WRITE_LOCKED | myd->tid)) abortTx(myd);
    }
    myd->writeSet.addData(vraddr, length);
}

// Same as checkRange(), but handles a range and is not inlined (slow-path)
static void checkRangeSlow(OpData* const myd, std::atomic<uint64_t>* mBeg, std::atomic<uint64_t>* mEnd) {
    // Handle wrap-around of lock ranges
    if (mEnd < mBeg) {
        if (myd->tx_type == TX_IS_UPDATE) myd->readSet.add(&gHashLock[0], mBeg);
        for (std::atomic<uint64_t>* mutex = &gHashLock[0]; mutex <= mBeg; mutex++) {
            uint64_t sl = mutex->load(std::memory_order_acquire);
            if (!isUnlockedOrLockedByMe(myd->rClock, myd->tid, sl)) abortTx(myd);
        }
        mBeg = mEnd;
        mEnd = &gHashLock[NUM_LOCKS-1];
    }
    // When in a write tx, loads must be added to the read-set
    if (myd->tx_type == TX_IS_UPDATE) myd->readSet.add(mBeg, mEnd);
    for (std::atomic<uint64_t>* mutex = mBeg; mutex <= mEnd; mutex++) {
        uint64_t sl = mutex->load(std::memory_order_acquire);
        if (!isUnlockedOrLockedByMe(myd->rClock, myd->tid, sl)) abortTx(myd);
    }
}

// Helper function to check we're writing in 'main'. Used by pload() and string utils.
// Make sure to issue a "asm volatile ("" : : : "memory")" before calling this.
inline static void checkRange(OpData* const myd, void* mainaddr, std::size_t length) {
    if (myd == nullptr || mainaddr < BEGIN_MAIN_REGION || mainaddr >= END_MAIN_REGION) return;
    //assert(myd ->tx_type != TX_IS_IRR_READ);//TODO: remove 
    std::atomic<uint64_t>* mBeg  = &gHashLock[hidx(mainaddr)];
    std::atomic<uint64_t>* mEnd  = &gHashLock[hidx((uint8_t*)mainaddr + length-1)];
    if (mBeg == mEnd) {
        // Fast path for single-lock checks
        if (myd->tx_type == TX_IS_UPDATE) myd->readSet.add(mBeg, mEnd);
        uint64_t sl = mBeg->load(std::memory_order_acquire);
        if (!isUnlockedOrLockedByMe(myd->rClock, myd->tid, sl)) abortTx(myd);
    } else {
        // Slow path for multi-lock checks
        checkRangeSlow(myd, mBeg, mEnd);
    }
}

// T is typically a pointer to a node, but it can be integers or other stuff, as long as it fits in 64 bits
template<typename T> struct persist {
    T vrmain;

    persist() { }

    persist(T initVal) { pstore(initVal); }

    // Casting operator
    operator T() { return pload(); }
    // Casting to const
    operator T() const { return pload(); }

    // Prefix increment operator: ++x
    void operator++ () { pstore(pload()+1); }
    // Prefix decrement operator: --x
    void operator-- () { pstore(pload()-1); }
    void operator++ (int) { pstore(pload()+1); }
    void operator-- (int) { pstore(pload()-1); }
    persist<T>& operator+= (const T& rhs) { pstore(pload() + rhs); return *this; }
    persist<T>& operator-= (const T& rhs) { pstore(pload() - rhs); return *this; }

    // Equals operator
    template <typename Y, typename = typename std::enable_if<std::is_convertible<Y, T>::value>::type>
    bool operator == (const persist<Y> &rhs) { return pload() == rhs; }
    // Difference operator: first downcast to T and then compare
    template <typename Y, typename = typename std::enable_if<std::is_convertible<Y, T>::value>::type>
    bool operator != (const persist<Y> &rhs) { return pload() != rhs; }
    // Relational operators
    template <typename Y, typename = typename std::enable_if<std::is_convertible<Y, T>::value>::type>
    bool operator < (const persist<Y> &rhs) { return pload() < rhs; }
    template <typename Y, typename = typename std::enable_if<std::is_convertible<Y, T>::value>::type>
    bool operator > (const persist<Y> &rhs) { return pload() > rhs; }
    template <typename Y, typename = typename std::enable_if<std::is_convertible<Y, T>::value>::type>
    bool operator <= (const persist<Y> &rhs) { return pload() <= rhs; }
    template <typename Y, typename = typename std::enable_if<std::is_convertible<Y, T>::value>::type>
    bool operator >= (const persist<Y> &rhs) { return pload() >= rhs; }

    // Operator arrow ->
    T operator->() { return pload(); }

    // Copy constructor
    persist<T>(const persist<T>& other) { pstore(other.pload()); }

    // Assignment operator from a persist<T>
    persist<T>& operator=(const persist<T>& other) {
        pstore(other.pload());
        return *this;
    }

    // Assignment operator from a value
    persist<T>& operator=(T value) {
        pstore(value);
        return *this;
    }

    // Operator &
    T* operator&() {
        return (T*)this;
    }

    // Store interposing: acquire the lock and write in 'main'
    inline void pstore(T newVal) {
        uint8_t* vraddr = (uint8_t*)&vrmain;
        OpData* const myd = tl_opdata;
        // We don't acquire locks for data outside PM (that woud make more overhead on the logging system)
        if (myd != nullptr && vraddr >= BEGIN_MAIN_REGION && vraddr < END_MAIN_REGION) {
            // Logs stores and acquires locks, or aborts
            logLockRange(vraddr, sizeof(T));
        }
        vrmain = newVal;
    }

    // This is similar to an undo-log load interposing: do a single post-check
    // TODO: This is not ARM/PPC compatible code. Copy code from TL2Undo instead
    // TODO: try to specialize based on enable_if() https://en.cppreference.com/w/cpp/types/enable_if
    inline T pload() const {
        if(tl_opdata == nullptr) return vrmain; // TODO: is there a more elegant option?
	if(tl_opdata->tx_type == TX_IS_IRR_READ){
		T lval = *(T*)((uint8_t*)&vrmain + MAIN_REGION_SIZE);
		return lval;
	}
	// Else: read / update tx ; reading from main region
	T lval = vrmain;
        asm volatile ("" : : : "memory");
        checkRange(tl_opdata, (void*)&vrmain, sizeof(T)); // Aborts if lock is inconsistent or taken
        return lval;
    }
};


class PTM;
extern PTM gTL4x;


class PTM {
private:
    // Padding on x86 should be on 2 cache lines
    static const int                       CLPAD = 128/sizeof(uintptr_t);
    bool                                   reuseRegion {false};                 // used by the constructor and initialization
    int                                    pfd {-1};
    int                                    vfd {-1};
    alignas(128) OpData                   *opDesc;
    EsLoco2<persist>                       esloco {};
    std::thread                            copyThread;
    std::atomic<bool>                      destructorCalled {false};

public:
    URCU                                   urcu {};
    struct tmbase : public tl4x::tmbase { };

    PTM() {
    	assert(sizeof(PMetadata)%64 == 0);
    	gHashLock = new std::atomic<uint64_t>[NUM_LOCKS];
    	for (uint64_t i = 0; i < NUM_LOCKS; i++) gHashLock[i].store(0, std::memory_order_relaxed);
        gSeqArray = new MainBackSeq[NUM_LOCKS];
        std::memset(gSeqArray, 0, NUM_LOCKS*sizeof(MainBackSeq));
        opDesc = new OpData[REGISTRY_MAX_THREADS];
        for (uint64_t it=0; it < REGISTRY_MAX_THREADS; it++) {
            opDesc[it].tid = it;
            opDesc[it].myrand = (it+1)*12345678901234567ULL;
        }
        mapPersistentRegion(PM_FILE_NAME, (uint8_t*)PM_REGION_BEGIN, PM_REGION_SIZE);
        // If recovery is needed, it will be triggered inside mapVolatileRegion()
        mapVolatileRegion(VR_FILE_NAME, VR_REGION_ADDR, PM_REGION_SIZE);
        copyThread = std::thread([this]() { copyThreadLoop(); });
    }

    ~PTM() {
        destructorCalled = true;
        copyThread.join();
        uint64_t totalAborts = 0;
        uint64_t totalCommits = 0;
        for (int it=0; it < REGISTRY_MAX_THREADS; it++) {
            totalAborts += opDesc[it].numAborts;
            totalCommits += opDesc[it].numCommits;
        }
        printf("totalAborts=%ld  totalCommits=%ld  abortRatio=%.1f%%   usedPM=%ld MB\n",
                totalAborts, totalCommits, 100.*totalAborts/(1+totalCommits), (esloco.getUsedSize()*64)/(24*1024*1024));
        delete[] opDesc;
        delete[] gHashLock;
        delete[] gSeqArray;
    }

    static std::string className() { return "TL4x"; }

    void mapPersistentRegion(const char* filename, uint8_t* regionAddr, const uint64_t regionSize) {
        // Check if the file already exists or not
        struct stat buf;
        if (stat(filename, &buf) == 0) {
            // File exists
            pfd = open(filename, O_RDWR|O_CREAT, 0755);
            if (pfd < 0) {
                perror("open() error");
                assert(pfd >= 0);
            }
            reuseRegion = true;
        } else {
            // File doesn't exist
            pfd = open(filename, O_RDWR|O_CREAT, 0755);
            if (pfd < 0) {
                perror("open() error");
                assert(pfd >= 0);
            }
            if (lseek(pfd, regionSize-1, SEEK_SET) == -1) {
                perror("lseek() error");
            }
            if (write(pfd, "", 1) == -1) {
                perror("write() error");
            }
        }
        // Try one time to mmap() with DAX. If it fails due to MAP_FAILED, then retry without DAX.
        // We may fail because the address is not available. Retry at most 4 times, then give up.
        uint64_t dax_flag = MAP_SYNC;
        // Don't try to enable DAX if we're doing disk persistence
#ifdef PWB_IS_DISK
        dax_flag = 0;
#endif
        void* got_addr = NULL;
        for (int i = 0; i < 4; i++) {
            got_addr = mmap(regionAddr, regionSize, (PROT_READ | PROT_WRITE), MAP_SHARED_VALIDATE | dax_flag, pfd, 0);
            if (got_addr == regionAddr) {
#ifndef PWB_IS_DISK
                if (dax_flag == 0) printf("WARNING: running with DAX disabled\n");
#endif
                break;
            }
            if (got_addr == MAP_FAILED) {
                // Failed to mmap(). Let's try without DAX.
                dax_flag = 0;
            } else {
                // mmap() worked but in wrong address. Let's unmap() and try again.
                munmap(got_addr, regionSize);
                // Sleep for a couple seconds before retrying the mmap()
                usleep(2000);
            }
        }
        if (got_addr == MAP_FAILED) {
            perror("ERROR: mmap() returned MAP_FAILED !!! ");
            assert(false);
        }
        if (got_addr != regionAddr) {
            printf("got_addr = %p instead of %p\n", got_addr, regionAddr);
            perror("Retry mmap() in a couple of seconds");
            std::exit(42);
        }
        // Check if the header is consistent and only then can we attempt to re-use, otherwise we clear everything that's there
        //if (reuseRegion) reuseRegion = (pmd->id == PMetadata::MAGIC_ID);
    }

    // Maps the volatile memory region
    void mapVolatileRegion(const char* filename, uint8_t* regionAddr, const uint64_t regionSize) {
        // Check if the file already exists or not
        struct stat buf;
        if (stat(filename, &buf) == 0) {
            // File exists
            vfd = open(filename, O_RDWR|O_CREAT, 0755);
            if (vfd < 0) {
                perror("open() error");
                assert(vfd >= 0);
            }
        } else {
            // File doesn't exist
            vfd = open(filename, O_RDWR|O_CREAT, 0755);
            if (vfd < 0) {
                perror("open() error");
                assert(vfd >= 0);
            }
            if (lseek(vfd, regionSize-1, SEEK_SET) == -1) {
                perror("lseek() error");
            }
            if (write(vfd, "", 1) == -1) {
                perror("write() error");
            }
        }
        // mmap() volatile memory range 'main'
        uint8_t* got_addr = (uint8_t *)mmap(regionAddr, regionSize, (PROT_READ | PROT_WRITE), MAP_SHARED, vfd, 0);
        if (got_addr == MAP_FAILED) {
            perror("ERROR: mmap() returned MAP_FAILED !!! ");
            assert(false);
        }
        if (got_addr != regionAddr) {
            printf("got_addr = %p instead of %p\n", got_addr, regionAddr);
            perror("Retry mmap() in a couple of seconds");
            std::exit(42);
        }
        // If the file has just been created or if the header is not consistent, clear everything.
        // Otherwise, re-use and recover to a consistent state.
        if (reuseRegion) {
            recover();
            // Initialize the volatile data of the EsLoco allocator
            readTx([&] () {
                esloco.init(regionAddr+sizeof(PMetadata), regionSize/2 - sizeof(PMetadata), false);
            });
        } else {
            // Call PMedata constructor
            new (pmd) PMetadata();
            // We reserve half of the volatile mmap region for 'main' and allocate only there
            updateTx([&] () {
                // Initialize the allocator and allocate an array for the root pointers
                esloco.init(regionAddr+sizeof(PMetadata), regionSize/2 - sizeof(PMetadata), true);
                pmd->root = esloco.malloc(sizeof(persist<void*>)*MAX_ROOT_POINTERS);
            });
            pmd->id = PMetadata::MAGIC_ID;
            // The 'root' and 'id' are not 'persist' therefore, we need to manually duplicate them on 'back'
            std::memcpy(BEGIN_BACK_REGION+offsetof(PMetadata, root), BEGIN_MAIN_REGION+offsetof(PMetadata, root), sizeof(PMetadata::root));
            std::memcpy(BEGIN_BACK_REGION+offsetof(PMetadata, id), BEGIN_MAIN_REGION+offsetof(PMetadata, id), sizeof(PMetadata::id));
        }
    }


    void waitForPersist() {
	    while (true){
	    	uint64_t counter = gCounter.load();
		    if((counter & COUNTER_ON) == COUNTER_ON){
			    if(gCounter.compare_exchange_strong(counter,counter+1ULL)) break;
		    }
		    std::this_thread::yield();
	    }

	    while (true) {
	    	uint64_t st = gState.load();
		    if (st == STATE_READ_SNAPSHOT) break;
		    std::this_thread::yield(); 
	    }
    }
		

    inline void beginTx(OpData* myd, const int tid) {
        // Clear the logs of the previous transaction
        myd->writeSet.reset();
        myd->readSet.reset();
        if (myd->tx_type == TX_IS_IRR_READ){
            waitForPersist();
            return;
        }
        myd->rClock = gClock.load(); // This is GVRead()
        // No need to arrive on the URCU if we're in a read-only transaction
        if (myd->tx_type == TX_IS_UPDATE) {
        	urcu.arrive(tid);
        	myd->state = gState.load();
        }
    }

    // This PTM does eager locking, which means that by now all locks have been acquired.
    inline bool endTx(OpData* myd, const int tid) {
        if(myd->tx_type == TX_IS_IRR_READ){
		    gCounter.fetch_sub(1);
		    myd->numCommits++;
		    myd->attempt = 0;
		    return true;
	    }
	// Check if this is a read-only transaction and if so, commit immediately
        if (myd->writeSet.numEntries == 0) {
            myd->attempt = 0;
            if(myd->tx_type == TX_IS_UPDATE) urcu.depart(tid);
            return true;
        }
        // Validate the read-set
        if (!myd->readSet.validate(myd->rClock, tid)) abortTx(myd);
        // Tx is now committed and holding all the locks.
        // Increment the global clock on TL2
        uint64_t nextClock = gClock.fetch_add(1)+1;
        // Read the current state to see if we need to duplicate the changes, or the blocks or neither
        const uint8_t state = myd->state;
        if (state == STATE_DUPLICATE) {
            myd->writeSet.duplicateChangesOnBack();
        }/*
        else if (state == STATE_DUPLICATE_BLOCK) {
            myd->writeSet.duplicateBlocksOnBack();
        }*/
         // else, we're in PERSIST and we don't touch 'back'
        // Unlock and set new sequence on the locks and on gSeqArray
        myd->writeSet.unlock(nextClock, state, true, tid);
        urcu.depart(tid);
        myd->numCommits++;
        myd->attempt = 0;
        return true;
    }

    template<typename F> void transaction(F&& func, int txType=TX_IS_UPDATE) {
        if (tl_opdata != nullptr) {
            func();
            return ;
        }
        const int tid = ThreadRegistry::getTID();
        OpData* myd = &opDesc[tid];
        tl_opdata = myd;
        myd->tx_type = txType;
        setjmp(myd->env);
        myd->attempt++;
        backoff(myd, myd->attempt);
        beginTx(myd, tid);
        func();
        endTx(myd, tid);
        tl_opdata = nullptr;
    }

    // It's silly that these have to be static, but we need them for the (SPS) benchmarks due to templatization
    template<typename F> static void updateTx(F&& func) { gTL4x.transaction(func, TX_IS_UPDATE); }
    template<typename F> static void readTx(F&& func) { gTL4x.transaction(func, TX_IS_READ); }
    template<typename F> static void irrevocableReadTx(F&& func) { gTL4x.transaction(func, TX_IS_IRR_READ); }
    // There are no sequential durable transactions in TL2 but we "emulate it" with a concurrent+durable tx
    template<typename F> static void updateTxSeq(F&& func) { gTL4x.transaction(func, TX_IS_UPDATE); }
    template<typename F> static void readTxSeq(F&& func) { gTL4x.transaction(func, TX_IS_READ); }
    template<typename F> static void irrevocableReadTxSeq(F&& func) { gTL4x.transaction(func, TX_IS_IRR_READ); }
    // Recover 'main' and 'back' from the p0/p1 replica with the highest timestamp
    void recover() {
        printf("Executing recovery...\n");
        // Decide what is the persistent region we should recover from, based on the timestamp of each region
        PMetadata* p0 = (PMetadata*)BEGIN_P0_REGION;
        PMetadata* p1 = (PMetadata*)BEGIN_P1_REGION;
        PMetadata* pconsistent = (p1->p_timestamp > p0->p_timestamp) ? p1 : p0;
        PMetadata* pinconsistent = (p1->p_timestamp > p0->p_timestamp) ? p0 : p1;
        if (pconsistent->id != PMetadata::MAGIC_ID) {
            printf("Expected id to be 0x%lx but it was 0x%lx\n", PMetadata::MAGIC_ID, pconsistent->id);
            assert(false);
        }
        printf("recover:  p0=%ld   p1=%ld\n", p0->p_timestamp, p1->p_timestamp);
        // Copy all data in the persistent region to 'main' and 'back' and the opposite persistent region
        std::memcpy(BEGIN_BACK_REGION, pconsistent, PM_REGION_SIZE/2);
        std::memcpy(BEGIN_MAIN_REGION, pconsistent, PM_REGION_SIZE/2);
        // Don't copy the timestamp, which is the first word in the PMetadata
        std::memcpy((uint8_t*)pinconsistent+sizeof(uint64_t), (uint8_t*)pconsistent+sizeof(uint64_t), PM_REGION_SIZE/2 - sizeof(uint64_t));
        // Flush the persistent memory with clwbs
        for (uint64_t i = 0; i < PM_REGION_SIZE/2; i+= 64) PWB((uint8_t*)pinconsistent + i);
        PFENCE();
        // The following macro does nothing in PM
        MSYNC((uint8_t*)pinconsistent+sizeof(uint64_t), PM_REGION_SIZE/2 - sizeof(uint64_t), MS_SYNC);
        // Now set the new timestamp
        pinconsistent->p_timestamp = pconsistent->p_timestamp;
        PWB(&pinconsistent->p_timestamp);
        PSYNC();
        MSYNC(pinconsistent, sizeof(uint64_t), MS_SYNC);
        // Set the global clock of TL2 to be the same as the current replica's timestamp+1
        gClock.store(pconsistent->p_timestamp+1);
        printf("Recovery complete!\n");
    }

    // Random number generator used by the backoff scheme
    inline uint64_t marsagliaXORV(uint64_t x) {
        if (x == 0) x = 1;
        x ^= x << 6;
        x ^= x >> 21;
        x ^= x << 7;
        return x;
    }

    // Backoff for a random amount of steps in the range [16, 16*attempt]. Inspired by TL2
    inline void backoff(OpData* myd, uint64_t attempt) {
        if (attempt < 2) return;
        if (attempt == 100000) printf("Ooops, looks like we're stuck attempt=%ld\n", attempt);
        myd->myrand = marsagliaXORV(myd->myrand);
        uint64_t stall = myd->myrand & 0xF;
        stall += attempt >> 2;
        stall *= 8;
        std::atomic<uint64_t> iter {0};
        while (iter.load() < stall) iter.store(iter.load()+1);
    }

    template<typename R,class F> inline static R readTx(F&& func) {
        gTL4x.transaction([&]() {func();}, TX_IS_READ);
        return R{};
    }
    template<typename R,class F> inline static R updateTx(F&& func) {
        gTL4x.transaction([&]() {func();}, TX_IS_UPDATE);
        return R{};
    }
    template<typename R,class F> inline static R irrevocableReadTx(F&& func) {
	    gTL4x.transaction([&]() {func();}, TX_IS_IRR_READ);
	    return R{};
    }
    template <typename T, typename... Args> static T* tmNew(Args&&... args) {
        if (tl_opdata == nullptr) {
            printf("ERROR: Can not allocate outside a transaction\n");
            return nullptr;
        }
        void* ptr = gTL4x.esloco.malloc(sizeof(T));
        // If we get nullptr then we've ran out of PM space
        assert(ptr != nullptr);
        // If there is an abort during the 'new placement', the transactions rolls
        // back and will automatically revert the changes in the allocator metadata.
        new (ptr) T(std::forward<Args>(args)...);  // new placement
        return (T*)ptr;
    }

    template<typename T> static void tmDelete(T* obj) {
        if (obj == nullptr) return;
        if (tl_opdata == nullptr) {
            printf("ERROR: Can not de-allocate outside a transaction\n");
            return;
        }
        obj->~T();
        gTL4x.esloco.free(obj);
    }

    static void* tmMalloc(size_t size) {
        if (tl_opdata == nullptr) {
            printf("ERROR: Can not allocate outside a transaction\n");
            return nullptr;
        }
        void* obj = gTL4x.esloco.malloc(size);
        return obj;
    }

    static void tmFree(void* obj) {
        if (obj == nullptr) return;
        if (tl_opdata == nullptr) {
            printf("ERROR: Can not de-allocate outside a transaction\n");
            return;
        }
        gTL4x.esloco.free(obj);
    }

    static void* pmalloc(size_t size) {
        if (tl_opdata == nullptr) {
            printf("ERROR: Can not allocate outside a transaction\n");
            return nullptr;
        }
        return gTL4x.esloco.malloc(size);
    }

    static void pfree(void* obj) {
        if (obj == nullptr) return;
        if (tl_opdata == nullptr) {
            printf("ERROR: Can not de-allocate outside a transaction\n");
            return;
        }
        gTL4x.esloco.free(obj);
    }

    static void* tmMemcpy(void* dst, const void* src, std::size_t count) {
        void* result = nullptr;
        OpData* const myd = tl_opdata;
        if (myd != nullptr) {
            logLockRange(dst, count);   // Aborts if 'dst' is already locked
            result = std::memcpy(dst, src, count);
            asm volatile ("" : : : "memory");
            checkRange(myd, (void*)src, count);  // Aborts if 'src' is locked or was modified during tx
        } else {
            result = std::memcpy(dst, src, count);
        }
        return result;
    }

    static int tmMemcmp(const void* lhs, const void* rhs, std::size_t count) {
        int result = std::memcmp(lhs, rhs, count);
        asm volatile ("" : : : "memory");
        OpData* const myd = tl_opdata;
        if (myd != nullptr) {
            checkRange(myd, (void*)lhs, count);
            checkRange(myd, (void*)rhs, count);
        }
        return result;
    }

    static int tmStrcmp(const char* lhs, const char* rhs, std::size_t count) {
        int result = std::strncmp(lhs, rhs, count);
        if (tl_opdata != nullptr) {
            if (tl_opdata->tx_type == TX_IS_IRR_READ) return result;
        }
        asm volatile ("" : : : "memory");
        OpData* const myd = tl_opdata;
        if (myd != nullptr) {
            checkRange(myd, (void*)lhs, count);
            checkRange(myd, (void*)rhs, count);
        }
        return result;
    }

    static void* tmMemset(void* dst, int ch, std::size_t count) {
        if (tl_opdata != nullptr) logLockRange(dst, count);   // Aborts if 'dst' is already locked
        return std::memset(dst, ch, count);
    }

    // Get a root pointer
    static inline void* get_object(int idx) {
        return ((persist<void*>*)pmd->root)[idx].pload();
    }

    // Set a root pointer
    static inline void put_object(int idx, void* obj) {
        ((persist<void*>*)pmd->root)[idx].pstore(obj);
    }

private:
    // Loop executed by the copy-thread
    void copyThreadLoop() {
        uint64_t lastSyncedTxn = 0;
	    bool doPersist = false;
	    uint64_t persistTime = 0;
	    uint64_t timesSkipped = 0;
	    uint64_t timesChecked = 0;
        // Loop until we call the destructor of the PTM
        while (!destructorCalled) {
            // Loop until enough transactions have passed that we need to start a sync
	        // or gCounter is not empty
            if ((lastSyncedTxn != 0 && gClock.load() < (lastSyncedTxn + MIN_TXN_BETWEEN_SYNC))) {
                if(gCounter.load() == COUNTER_ON ){
                    std::this_thread::yield();
                    timesSkipped++;
                    continue;
                }
            }else{
                doPersist = true;
            }

            timesChecked++;
            auto beginCopyTime = std::chrono::steady_clock::now();
            if(doPersist){
                // Start a SYNC operation
                if (DEBUG_SYNC) printf("copy-thread starting sync...\n");

                const uint64_t txnBeforeCopy = gClock.load();
                lastSyncedTxn = txnBeforeCopy;
            }
            //set counter off
            gCounter.fetch_sub(COUNTER_ON);
            gState.store(STATE_PERSIST);
	        urcu.synchronize();
	        gState.store(STATE_READ_SNAPSHOT);
#ifndef TL4X_NO_SYNC
	        if(doPersist){
		        // Decide what is the address of the persistent region we should use based on the timestamp of each region
		        PMetadata* p0 = (PMetadata*)BEGIN_P0_REGION;
		        PMetadata* p1 = (PMetadata*)BEGIN_P1_REGION;
		        PMetadata* pm = (p1->p_timestamp > p0->p_timestamp) ? p0 : p1;
		        const uint64_t usedSize = esloco.getUsedSize()+sizeof(PMetadata);
		        if (DEBUG_SYNC) printf("usedSize = %ld MB  %ld\n", usedSize/(1024*1024), sizeof(EsLoco2<persist>::ELMetadata));
		        // Copy data from 'back' to persistent replica. Will copy only the data modified from the last copy.
		        copyBackToP01(pm, usedSize, lastSyncedTxn);
		        // Ensure the writes on the persistent replica are durable before setting the timestamp
		        PFENCE();
		        // This macro does nothing for PM, it's used only for disk persistence
		        MSYNC((uint8_t*)pm+sizeof(uint64_t), usedSize, MS_SYNC);
		        pm->p_timestamp = lastSyncedTxn;
		        PWB(&pm->p_timestamp);
		        PSYNC();
		        // This macro does nothing for PM, it's used only for disk persistence
		        MSYNC(pm, sizeof(uint64_t), MS_SYNC);
		        auto endMes = std::chrono::steady_clock::now();
		        persistTime =  (endMes - beginCopyTime).count() / 1000000ULL;
	        }
#endif
	        //assert(GETSTATE(gDupState.load()) == STATE_READ_SNAPSHOT);
            //compareTwoRegions((uint8_t*)pm+sizeof(PMetadata), BEGIN_BACK_REGION+sizeof(PMetadata), (PM_REGION_SIZE/2) -sizeof(PMetadata));
	        while (gCounter.load()>0){
	    	    /*std::cout << gState.load() << std::endl*/; // wait for readers to complete
	    	    std::this_thread::yield();
	        }
	        // Let the write-txn duplicate their writes again, this time the full block
	        gState.store(STATE_DUPLICATE_BLOCK);
	        gCounter.store(COUNTER_ON);
	        //assert(GETSTATE(gDupState.load()) == STATE_DUPLICATE_BLOCK);
            // Quiesce, thus guaranteeing that after the urcu.synchronize, all in-flight write-txns will be seeing the DUPLICATE_BLOCK state
	        urcu.synchronize();
            // Now copy all the blocks that changed in the mean time, from main to back.
            copyMainToBack();
            // Go back to the original state where write-txn duplicate only the modifications, no need to quiesce this time.
	        gState.store(STATE_DUPLICATE);

            // Account time spent on SYNC operation
            auto endCopyTime = std::chrono::steady_clock::now();
            uint64_t copyDuration = (endCopyTime-beginCopyTime).count() / 1000000ULL;
            //if (copyDuration >= 1000) printf("copy-thread took %ld seconds to execute a SYNC\n", copyDuration);
            if (DEBUG_SYNC) printf("copy-thread finished sync...\n");
            doPersist = false;
            //printf("Copy-thread spent %ld ms in a state cycle, %ld ms to persist. Cheched %ld, Skipped %ld\n", copyDuration, persistTime, timesChecked, timesSkipped);
        }
#if defined TL4x_MEASURE_TIME
	printf("Copy-thread spent %ld ms in a state cycle, %ld ms to persist. Cheched %ld, Skipped %ld\n", copyDuration, persistTime, timesChecked, timesSkipped);
#endif
    }

    // Helper function for copy-thread. Copies data from 'back' to the oldest persistent replica
    void copyBackToP01(PMetadata* pm, const uint64_t usedSize, uint64_t& lastSyncedTxn) {
        const uint64_t dataSize = (1 << BIT_SHIFT_HASH);
        const uint64_t pitimestamp = pm->p_timestamp;
        uint8_t* pbegin = (uint8_t*)pm;
        // Copy the PMetadata header but not the timestamp
        std::memcpy(pbegin+sizeof(uint64_t), BEGIN_BACK_REGION+sizeof(uint64_t), sizeof(PMetadata)-sizeof(uint64_t));
        // TODO: this assumes we're allocating less than what is protected by the array of locks, i.e. no wraparound. FIX ME!!!
        assert (usedSize/dataSize < NUM_LOCKS);
        for (uint64_t offset = sizeof(PMetadata); offset <= usedSize; offset += dataSize) {
            int64_t imut = hidx(BEGIN_MAIN_REGION+offset);
            if (gSeqArray[imut].back <= pitimestamp) continue;
            // The 'lastSyncedTxn' is max of TL2 clock before starting the procedure and the highest seen txn sequence
            if (gSeqArray[imut].back > lastSyncedTxn) lastSyncedTxn = gSeqArray[imut].back;
            // Copy from 'back' to the persistent region, one data block at a time
            std::memcpy(pbegin+offset, BEGIN_BACK_REGION+offset, dataSize);
            PWB(pbegin+offset);
        }
    }

    // Helper function for copy-thread. Copies data from 'main' to 'back'.
    // We copy only the blocks of data whose timestamp in gHashSeq[] does not match gHashLock[].
    void copyMainToBack() {
        const uint64_t usedSize = esloco.getUsedSize()+sizeof(PMetadata);
        const uint64_t dataSize = (1 << BIT_SHIFT_HASH);
        // Copy the PMetadata header but not the timestamp. No need for locks.
        std::memcpy(BEGIN_BACK_REGION+sizeof(uint64_t), BEGIN_MAIN_REGION+sizeof(uint64_t), sizeof(PMetadata)-sizeof(uint64_t));
        // TODO: add support for wraparound
        for (uint64_t offset = sizeof(PMetadata); offset <= usedSize; offset += dataSize) {
            int64_t imut = hidx(BEGIN_MAIN_REGION+offset);
            uint64_t sl = gHashLock[imut].load();
            if (isLocked(sl) || (gSeqArray[imut].main == gSeqArray[imut].back)) continue;
            // This block of data may have been modified while we were in PERSIST state therefore,
            // we need to copy the latest from main. Acquire the lock before doing the copy.
            if (!gHashLock[imut].compare_exchange_strong(sl, WRITE_LOCKED|REGISTRY_MAX_THREADS)) continue;
            // Copy the dataSized block from 'main' to 'back'
            std::memcpy(BEGIN_BACK_REGION+offset, BEGIN_MAIN_REGION+offset, dataSize);
            // Set the new sequence for 'back' and release the lock with the same sequence (we didn't change anything in 'main')
            gSeqArray[imut].back = gSeqArray[imut].main;
            gHashLock[imut].store(sl, std::memory_order_release);
        }
    }

    // For debugging only
    void dumpMemoryRegion(const char* str, void* addr, size_t size) {
        printf("%s: ", str);
        for (int i = 0; i < size; i++) printf("%2x",*((uint8_t*)addr + i));
        printf("\n");
    }

    // For debugging only
    void compareTwoRegions(uint8_t* lhs, uint8_t* rhs, size_t size) {
        for (size_t ibyte = 0; ibyte < size; ibyte++) {
            if (lhs[ibyte] != rhs[ibyte]) {
                printf("Regions differ in byte %ld with %2x versus %2x  rhs address is %p\n", ibyte, lhs[ibyte], rhs[ibyte], &rhs[ibyte]);
                dumpMemoryRegion("main", BEGIN_MAIN_REGION, 512);
                dumpMemoryRegion("back", BEGIN_BACK_REGION, 512);
                assert(false);
            }
        }
    }
};


//
// Wrapper methods to the global TM instance. The user should use these:
//
template<typename R, typename F> static R updateTx(F&& func) { return gTL4x.transaction<R>(func, TX_IS_UPDATE); }
template<typename R, typename F> static R readTx(F&& func) { return gTL4x.transaction<R>(func, TX_IS_READ); }
template<typename R, typename F> static R irrevocableReadTx(F&& func) { return gTL4x.transaction<R>(func, TX_IS_IRR_READ); }
template<typename F> static void updateTx(F&& func) { gTL4x.transaction(func, TX_IS_UPDATE); }
template<typename F> static void readTx(F&& func) { gTL4x.transaction(func, TX_IS_READ); }
template<typename F> static void irrevocableReadTx(F&& func) { gTL4x.transaction(func, TX_IS_IRR_READ); }
template<typename T, typename... Args> T* tmNew(Args&&... args) { return PTM::tmNew<T>(args...); }
template<typename T> void tmDelete(T* obj) { PTM::tmDelete<T>(obj); }
static void* tmMalloc(size_t size) { return PTM::tmMalloc(size); }
static void tmFree(void* obj) { PTM::tmFree(obj); }


#ifndef INCLUDED_FROM_MULTIPLE_CPP
//
// Place these in a .cpp if you include this header from multiple files (compilation units)
//
// Global/singleton to hold all the thread registry functionality
ThreadRegistry gThreadRegistry {};
// Array of locks
std::atomic<uint64_t> *gHashLock {nullptr};
// Arrays of sequences for 'main' and 'back'
MainBackSeq *gSeqArray {nullptr};
// Global clock for TL2
alignas(128) std::atomic<uint64_t> gClockPaddingA {0};
alignas(128) std::atomic<uint64_t> gClock {1};
alignas(128) std::atomic<uint64_t> gClockPaddingB {0};



// Persistent state can be DUPLICATE or PERSIST, controlled by the copy-thread.
alignas(128) std::atomic<uint64_t> gState {STATE_DUPLICATE}; // TODO: works but not nice. Need to shift or init with another value
alignas(128) std::atomic<uint64_t> gCounter {COUNTER_ON};
alignas(128) std::atomic<uint64_t> gClockPaddingC {0};
// PTM singleton
PTM gTL4x {};
// Thread-local data of the current ongoing transaction
thread_local OpData* tl_opdata {nullptr};
// This is where every thread stores the tid it has been assigned when it calls getTID() for the first time.
// When the thread dies, the destructor of ThreadCheckInCheckOut will be called and de-register the thread.
thread_local ThreadCheckInCheckOut tl_tcico {};
// Helper function for thread de-registration
void thread_registry_deregister_thread(const int tid) {
    gThreadRegistry.deregister_thread(tid);
}
// This is called from persist::load()/store() and endTx() if the read-set validation fails.
[[noreturn]] __attribute__ ((noinline)) void abortTx(OpData* myd) {
    if (myd->tx_type == TX_IS_UPDATE) {
        myd->writeSet.rollbackMain(myd->tid);
        uint64_t nextClock = gClock.fetch_add(1)+1;
        // Unlock with the new sequence, setting the sequences on gSeqArray as needed
        myd->writeSet.unlock(nextClock, gState.load(), false, myd->tid);
        // After settings the sequences and releasing the locks we can finally depart from URCU
        gTL4x.urcu.depart(myd->tid);
    }
    myd->numAborts++;
    std::longjmp(myd->env, 1);
}
#endif // INCLUDED_FROM_MULTIPLE_CPP

}

